""" Tools for reading/writing function configuration. """
import re
import sys


# [arm_func|thumb_func] <address> [module] [name]
cfg_re = re.compile(r'(thumb_func|arm_func)?\s*(?:0x)?([0-9a-fA-F]{7,8})(?:\s+(\S+\.s))?(?:\s+(\S+)\r?\n)?')


def read_config(path):
    """ Reads a configuration file into an address map.

    Args:
        path (str): Path to function configuration file.

    Returns:
        dict: Mapping from function address to (name, module) tuples.
    """
    global cfg_re
    addr_map = {}
    with open(path, 'r') as f:
        for line in f:
            index = line.find('#')
            if index != -1:
                line = line[:index]
            m = cfg_re.match(line)
            if m:
                func_type, addr, module, name = m.groups()
                addr = int(addr, 16)
                name = name if name else None
                if func_type in (None, 'thumb_func'):
                    addr_map[addr] = name, module
    return addr_map


def strip_unknown(config):  # Strip unknown names from config
    unk_re = re.compile(r'sub_[0-9a-fA-F]{7,8}')
    new_config = {}
    for k, (name, module) in config.items():
        if name and unk_re.match(name):
            name = None
        new_config[k] = (name, module)
    return new_config


def write_config(addr_map, path):
    """ Writes an address map to a configuration file.

    Args:
        addr_map (dict): Mapping from function address to (name, module) tuples.
        path (str): Path to write.
    """
    named = unnamed = 0
    current_module = None
    seen = set()
    with open(path, 'w', buffering=1) as f:
        nfuncs = str(len(addr_map))
        s0 = f'# {nfuncs} functions, '
        s1 = '{} named, {} unnamed'
        f.write(s0+' '*len(s1.format(nfuncs, nfuncs)) + '\n')
        f.write('# [arm_func|thumb_func] <address> [module] [name]\n')
        for addr in sorted(addr_map):
            name, module = addr_map[addr]
            parts = [f'0x{addr:07X}']
            if module:
                parts.append(module)
                if module != current_module:
                    if current_module:
                        seen.add(current_module)
                    if module in seen:
                        print(f"Warning: {addr:08X}: Module '{module}' was already seen!", file=sys.stderr)
                    f.write(f'# {module}\n')
                    current_module = module
            if name:
                parts.append(name)
                named += 1
            else:
                unnamed += 1
            f.write(' '.join(parts) + '\n')
        f.seek(len(s0))
        f.write(s1.format(named, unnamed))



""" Luvdis CLI. """
import sys
import argparse

from luvdis import __doc__ as description
from luvdis.config import read_config
import luvdis.disassemble
from luvdis.disassemble import State, ROM, eprint, BASE_ADDRESS, _set_debug


def parse_int(n):
    return int(n, 0)


parser = argparse.ArgumentParser(prog='luvdis', formatter_class=argparse.RawDescriptionHelpFormatter,
                                 description=description)
parser.add_argument('rom', type=str,
                    help='Path to GBA ROM to disassemble')
parser.add_argument('-o', type=str, dest='out', default=None,
                    help='Disassembly output path')
parser.add_argument('-c', '--config', type=str, dest='config', default=None,
                    help='Path to function configuration file')
parser.add_argument('-co', '--config_out', type=str, dest='config_out', default=None,
                    help='Optional function configuration output')
parser.add_argument('-D', '--debug', action='store_true', dest='debug',
                    help='Set debugging flag')
parser.add_argument('--min_calls', type=int, default=2,
                    help='Minimum number of calls required to consider a potential function. Default 2.')
parser.add_argument('--min_length', type=int, default=3,
                    help='Minimum codepath length required to consider a potential function. Default 3.')
parser.add_argument('--start', type=parse_int, default=BASE_ADDRESS,
                    help='Start address to disassemble. Default 0x8000000.')
parser.add_argument('--stop', type=parse_int, default=float('inf'),
                    help='Stop address to disassemble. Default infinity.')
parser.add_argument('--macros', type=str, default=None,
                    help='Optional path of macros to include. If omitted, embeds macros into the output file(s).')
parser.add_argument('--no_guess', action='store_false', dest='guess',
                    help='Disable function discovery & guessing entirely. Use only functions provided via -c.')


def main(args=None):
    args = args if args else sys.argv[1:]
    parsed = parser.parse_args(args)
    _set_debug(parsed.debug)
    functions = read_config(parsed.config) if parsed.config else None
    state = State(functions, parsed.min_calls, parsed.min_length, parsed.start, parsed.stop, parsed.macros)
    rom = ROM(parsed.rom)
    state.analyze_rom(rom, parsed.guess)
    if parsed.out is None:
        eprint(f'No output file specified. Printing to stdout.')
    state.dump(rom, parsed.out, parsed.config_out)


if __name__ == '__main__':
    main()



import math
import re

char_to_byte = {'あ': 0x01, 'い': 0x02, 'え': 0x04, '6': 0xA7,
                'A': 0xBB, 'B': 0xBC, 'C': 0xBD, 'D': 0xBE, 'E': 0xBF, 'F': 0xC0, 'G': 0xC1, 'H': 0xC2, 'I': 0xC3,
                'J': 0xC4, 'K': 0xC5, 'L': 0xC6, 'M': 0xC7, 'N': 0xC8, 'O': 0xC9, 'P': 0xCA, 'Q': 0xCB, 'R': 0xCC,
                'S': 0xCD, 'T': 0xCE, 'U': 0xCF, 'V': 0xD0, 'W': 0xD1, 'X': 0xD2, 'Y': 0xD3, 'Z': 0xD4, 'a': 0xD5,
                'b': 0xD6, 'c': 0xD7, 'd': 0xD8, 'e': 0xD9, 'f': 0xDA, 'g': 0xDB, 'h': 0xDC, 'i': 0xDD, 'j': 0xDE,
                'k': 0xDF, 'l': 0xE0, 'm': 0xE1, 'n': 0xE2, 'o': 0xE3, 'p': 0xE4, 'q': 0xE5, 'r': 0xE6, 's': 0xE7,
                't': 0xE8, 'u': 0xE9, 'v': 0xEA, 'w': 0xEB, 'x': 0xEC, 'y': 0xED, 'z': 0xEE, '\r': 0xFB, '\n': 0xFE,
                '.': 0xB8, '?': 0xAC, ' ': 0x00, '!': 0xAB, '^': 0xFA, ',': 0xB4, '\'': 0xB3, chr(0xfc): 0xFC,
                'さ': 0x0B}

string_re = re.compile(r'\.string\s+([\'"].*[\'"])')


def convert_match(m):
    s = eval(m.group(1))
    return convert(s)


def convert(s: str):
    b = [char_to_byte[c] for c in s] + [0xff]
    lines = []
    for i in range(math.ceil(len(b)/16)):
        row = b[16*i:16*(i+1)]
        line = '.byte ' + ', '.join(f'0x{x:02X}' for x in row)
        lines.append(line)
    return '\n'.join(lines)


def convert_file(path, out_path):
    with open(path, 'r') as f_in, open(out_path, 'w') as f_out:
        content = f_in.read()
        replaced, _ = string_re.subn(convert_match, content)
        f_out.write(replaced)


def convert_binary(s):
    b = bytes([char_to_byte[c] for c in s] + [0xff])
    with open('input.bin', 'wb') as f:
        f.write(b)


def convert_hex(h):
    with open('input.bin', 'wb') as f:
        f.write(bytearray.fromhex(h))



from setuptools import setup

from luvdis import __version__, url

with open('README.md', 'r') as f:
    long_description = f.read()


setup(name='Luvdis',
      version=__version__,
      description='Pure-Python Game Boy Advance Disassembler',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Ariel Antonitis',
      author_email='arant@mit.edu',
      url=url,
      packages=['luvdis', 'luvdis.test'],
      package_data={'luvdis': ['*.inc']},  # Include embedded functions
      include_package_data=True,
      entry_points={'console_scripts': ['luvdis = luvdis.__main__:main']},
      license='MIT',
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Console',
                   'Topic :: Software Development :: Disassemblers',
                   'Topic :: System :: Hardware',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: 3.8'],
      install_requires=[],
      python_requires='>=3.6')



import subprocess
import os
import sys

from os.path import join as joinp

from luvdis.charmap import convert_file

AS = joinp('asm', 'arm-none-eabi-as')
LD = joinp('asm', 'arm-none-eabi-ld')
LINK_TEMPLATE = joinp('asm', 'linker.ld')


def sremove(path):  # Silently remove file
    try:
        os.remove(path)
    except OSError:
        pass


def fill_link(template, addr, path, out_path):
    with open(template, 'r') as f:
        content = f.read()
    content = content.replace('%ADDR%', f'0x{addr:08x}')
    content = content.replace('%PATH%', path)
    with open(out_path, 'w') as f:
        f.write(content)


def assemble(path, addr=0, clean=True, debug=True):
    """ Assemble and link THUMB assembly code at an address.

    Args:
        path (str): Path to assemble.
        addr (int): Address to link at. If zero no linking is done.
        clean (bool): Whether to clean intermediate files. Defaults to True.
        debug (bool): Whether to export the binary as hex. Defaults to True.
    """
    root, ext = os.path.splitext(os.path.basename(path))
    asm_path = joinp('build', f'{root}.s')
    obj_path = joinp('build', f'{root}.o')
    link_path = joinp('build', f'{root}.ld')
    elf_path = joinp('build', f'{root}.elf')
    bin_path = joinp('build', f'{root}.bin')
    hex_path = joinp('build', f'{root}.hex')
    work_paths = asm_path, obj_path, link_path, elf_path
    print(f'AS  {root}')
    convert_file(path, asm_path)
    subprocess.run([AS, '-mcpu=arm7tdmi', '-o', obj_path, asm_path], check=True)  # Assemble
    if addr:  # Link
        print(f'LD  {root} at 0x{addr:08x}')
        fill_link(LINK_TEMPLATE, addr, obj_path, link_path)
        subprocess.run([LD, '-T', link_path, '-o', elf_path, obj_path], check=True)  # Link
    else:  # Just extract
        elf_path = obj_path
    subprocess.run(['arm-none-eabi-objcopy', '-O', 'binary', '-j', '.text', elf_path, bin_path], check=True)
    if debug:
        with open(bin_path, 'rb') as f_in, open(hex_path, 'w') as f_out:
            f_out.write(f_in.read().hex())
    if clean:
        for to_clean in work_paths:
            sremove(to_clean)
    size = os.path.getsize(bin_path)
    print(f'BIN {root} ({size} bytes)')


if __name__ == '__main__':
    args = sys.argv[1:]
    addr = int(args[1], base=16) if len(args) > 1 else 0
    assemble(args[0], addr)
